from flask import Flask


def create_app(config_file=None):
    """Factory to create the Flask application
    :param config_file: A python file or dict from which to load the config.
                        If omitted, the config file must be set using
                        the ``FRIDA_RPC_CONFIG`` environment variable.
                        If set, the environment variable is ignored
    :return: A `Flask` application instance
    """
    from .routes import frida_rpc_bp
    app = Flask('frida_rpc')
    _load_config(app, config_file)
    app.register_blueprint(frida_rpc_bp)
    return app


def _load_config(app, config_file):
    app.config.from_pyfile('defaults.cfg')
    if isinstance(config_file, dict):
        app.config.update(config_file)
    elif config_file:
        app.config.from_pyfile(config_file)
    else:
        app.config.from_envvar('FRIDA_RPC_CONFIG')



import logging
import frida
from .util import FridaServer

logging.getLogger(__name__).addHandler(logging.NullHandler())


class BaseOS:
    """This class abstracts some os and frida functions for getting informations about processes
     You can overide this class to implement custom and specified os functions"""

    def __init__(self, process, device='local', frida_location=None):

        self.process = process
        self.device = device
        self.frida = FridaServer(device_type=device, frida_location=frida_location)
        try:
            self.frida.get_processes()
        except frida.ServerNotRunningError:
            logging.critical('Server is not running try and start again')
            ret = self.frida.start_frida()
            if ret != 0:
                logging.error('Cannot start frida server  ')

    def spawn(self):
        self.frida_spawn()
        return True

    def frida_spawn(self):
        try:
            self.frida.frida_spawn(self.process)
        except frida.ExecutableNotFoundError:
            return False
        return True

    @property
    def is_app_installed(self):
        return True

    @property
    def is_app_running(self):
        processes = self.frida.get_processes()
        for process in processes:

            if process.name == self.process or process.pid == self.process:
                return True
        return False

    def run_rpc(self, *args):
        return self.frida.load_script(*args)



import json
from flask import current_app, Blueprint
from flask import request, jsonify, abort
from frida_rpc.lib import Command, InvalidDataException, CommandException

frida_rpc_bp = Blueprint('frida_rpc_bp', __name__)


@frida_rpc_bp.route('/rpc', methods=['GET', 'POST'])
def rpc_command():
    """run command on app request in format {'process':'process_name', 'script': base64=.script}"""
    request_json = request.get_json()
    app_debug = current_app.config.get('DEBUG')
    app_token = current_app.config.get('SECRET_KEY')
    req_token = request.values.get('SECRET_KEY')

    if (not app_debug and not req_token) or (not app_debug and app_token != req_token):
        abort(405, 'You must supply SECRET_KEY when not in debug')

    white_list = current_app.config.get('WHITE_LIST_APP', [])

    if request.values:
        c = Command(dict(request.values), whitelist=white_list)
    elif request_json:
        c = Command(json.loads(request_json))
    else:
        return abort(405)

    try:
        data = {'success': True}
        c.validate_data()
        message = c.run_command()
        data['result'] = message
        return jsonify(data)
    except InvalidDataException as e:
        return jsonify(e.args[0])
    except CommandException as e:
        return jsonify(e.args[0])



from ast import literal_eval as make_tuple
import frida
from . import lib


class FridaServer:
    default_location = '/data/local/tmp/system_server'

    def __init__(self, device_type='tcp', frida_location=None):
        self.device_type = device_type
        self.device = frida.get_device(device_type)
        self.frida_location = frida_location or self.default_location

    @staticmethod
    def start_frida():
        # p = subprocess.Popen(['su'], shell=False)
        # p.wait()
        # p = subprocess.Popen([self.frida_location, '&'], shell=False)
        # ret = p.wait()
        return True

    def get_processes(self):
        processes = self.device.enumerate_processes()

        return processes

    def get_apps(self):
        return self.device.enumerate_applications()

    def frida_spawn(self, process):
        pid = self.device.spawn([process])
        self.device.resume(pid)
        return pid

    def load_script(self, process, script, method_name, args):
        try:

            session = self.device.attach(process)
        except frida.ServerNotRunningError as e:
            raise lib.CommandException({'success': False, 'error': str(e)})
        except frida.ProcessNotFoundError as e:
            raise lib.CommandException({'success': False, 'error': str(e)})
        except frida.PermissionDeniedError as e:
            raise lib.CommandException({'success': False, 'error': str(e)})

        try:
            source = """rpc.exports = """ + script.decode() + """;"""
        except (UnicodeDecodeError, AttributeError):
            source = """rpc.exports = """ + script + """;"""

        try:
            script = session.create_script(source)
        except frida.InvalidArgumentError as e:
            raise lib.CommandException({'success': False, 'error': str(e), 'script': source})

        script.load()
        f_method = getattr(script.exports, method_name)
        try:
            result = f_method(*make_tuple(args))

            return result
        except frida.core.RPCException as e:
            raise lib.CommandException({'success': False, 'error': str(e), 'script': source})



import time
import base64
import binascii
import logging
from .system import BaseOS

logging.getLogger(__name__).addHandler(logging.NullHandler())


class InvalidDataException(Exception):
    pass


class CommandException(Exception):
    pass


class Command:
    """
    Implements a command interface for frida to the flask app .
    """
    ERROR = {'process_not_found': 'Please make sure process is running',
             'supply_process': 'Please supply process in command ', 'supply_script':
                 'Please supply script to process', 'app_not_whitelist': 'App not included in whiteliste contact admin',
             'app_not_installed': ' APP is not installed please contact admin',
             'unable_to_start_app': 'Unable to start app please check logs if error persist',
             'method_not_present': 'Method name must present', 'args_not_present': 'Args   must present'
             }

    def __init__(self, command_data, whitelist=None):
        """command_data = {'process':'com.google.chrome', 'script': 'javascript base64 code',
         'method_name':doCommand, 'args':[]}"""
        self.command_data = command_data
        self.system = BaseOS(command_data.get('process'), frida_location=command_data.get('frida_location'))
        self.WHITELIST_APP = whitelist or []
        self.app_args = self.command_data.get('args')
        self.process = self.command_data.get('process')
        self.method_name = self.command_data.get('method_name')
        self.script = self.command_data.get('script')

    def validate_data(self):

        data = {'success': False}
        if not self.command_data.get('process'):
            data['error'] = self.ERROR['supply_process']

        elif self.command_data.get('process') not in self.WHITELIST_APP:
            logging.critical(
                'Gotten process {} from command  which is not in whitelist '.format(self.command_data.get('process')))
            data['error'] = self.ERROR['app_not_whitelist']

        elif not self.command_data.get('script'):
            data['error'] = self.ERROR['supply_script']

        elif not self.command_data.get('method_name'):

            data['error'] = self.ERROR['method_not_present']

        elif not self.command_data.get('args'):

            data['error'] = self.ERROR['args_not_present']
        if data.get('error'):
            raise InvalidDataException(data)

        try:

            self.script = base64.b64decode(self.script, validate=True)
        except binascii.Error as e:
            logging.warning('An error occured while decoding script : {}'.format(str(e)))

    def run_command(self):
        data = {'success': False}
        process = self.command_data['process']
        logging.info('Gotten process {} from command '.format(process))
        logging.info('Check process {} is istalled '.format(process))
        if not self.system.is_app_installed:
            data['error'] = self.ERROR['app_not_installed']
            raise CommandException(data)
        logging.info('Check if process {} is running currently '.format(process))
        if not self.system.is_app_running:
            logging.warning('Process {} is not running '.format(process))
            logging.info('Spawning process {}  '.format(process))
            spawned = self.system.spawn()
            time.sleep(1)
            if spawned and self.system.is_app_running:
                logging.info('Successfully spawned process {}  and process is running '.format(process))

                return self.do_main_script()
            else:
                data['error'] = self.ERROR['unable_to_start_app']
                raise CommandException(data)
        else:
            logging.info('Process {} is running already '.format(process))
            return self.do_main_script()

    def do_main_script(self):
        result = self.system.run_rpc(self.process, self.script, self.method_name, self.app_args)
        if result == 'NOT_RUNNING':
            raise CommandException({'success': False, 'error': 'unable to connect to server please start frida'})
        return result



from setuptools import setup


def readme():
    with open('README.md') as f:
        return f.read()


setup(name='frida_rpc',
      version='0.1.1',
      description='Expose frida RPC as a service you can access over the web or network',

      long_description=readme(),
      long_description_content_type='text/markdown',
      classifiers=[

          'License :: OSI Approved :: MIT License',
          'Programming Language :: Python :: 3.6',
      ],
      keywords='frida rpc web network internet reverse engineering',
      url='https://github.com/ydaniels/frida-web-rpc',
      author='Yomi D',
      author_email='yomid4all@gmail.com',
      license='MIT',
      packages=['frida_rpc'],
      include_package_data=True,
      install_requires=[
          'flask',
          'frida'
      ],
      zip_safe=False)
